#include <stdio.h>
#include <stdarg.h>
#include <conio.h>
#include <bios.h>
#include <string.h>
#include <malloc.h>

#include "revive.h"

static void
show_map( void )
{
  int		i, j, k;
  int		cluster, sector;
  int		col;
  word		type;
  char		c;

  cput( 5, 1, START_LINE - 1, '', 0x03, 80 );
  csput( 40, START_LINE - 1, "%c0f Head 0 " );
  cput( 5, 1, START_LINE + Boot->Sectors, '', 0x03, 80 );
  csput( 40, START_LINE + Boot->Sectors, "%c0f Head 1 " );
  cput( 5, 1, START_LINE + 2 * Boot->Sectors + 1, '', 0x03, 80 );
  for ( i = 0; i < Tracks; i++ )
    for ( j = 0; j < Boot->HeadCnt; j++ )
      for ( k = 0; k < Boot->Sectors; k++ )
      {
	sector = ( i * Boot->HeadCnt + j ) * Boot->Sectors + k;
	if ( sector == 0 )
	  c = 'B';
	else if ( sector <= Boot->ResSecs - 1 )
	  c = 'r';
	else if ( sector <= Boot->ResSecs - 1 + Boot->FatSize )
	  c = 'F';
	else if ( sector <= Boot->ResSecs - 1 + Boot->FatSize * Boot->FatCnt )
	  c = 'f';
	else if ( sector <= Boot->ResSecs - 1 + Boot->FatSize * Boot->FatCnt + ( Boot->RootSiz >> 4 ) )
	  c = 'D';
	else
	{
	  sector -= Boot->ResSecs + Boot->FatSize * Boot->FatCnt +
		  ( Boot->RootSiz >> 4 );
	  cluster = 2 + sector / Boot->ClustSiz;
	  type = *( word * ) ( fat + cluster * 3 / 2 );
	  if ( cluster & 1 )
	    type >>= 4;
	  type &= 0xfff;
	  if ( type == 0 )
	    c = '.';
	  else if ( type >= 0xff0 && type < 0xff7 )
	    c = 'r';
	  else if ( type == 0xff7 )
	    c = 'X';
	  else
	    c = '';
	}
	if ( ( SECT_STATUS( i, j, k ) & 0xff00 ) == ERROR )
	  col = 0xa0;
	else
	  col = 0x20;
	cput( 4, i + 1, START_LINE + j * ( Boot->Sectors + 1 ) + k, c, col );
      }
}

static void
position_sector(  )
{
  int		i, j, k;
  int		sector, cluster;
  int		key;

  sector = 0;
  cput( 5, 1, START_LINE + 2 * ( Boot->Sectors + 1 ) + 1, ' ', 0x02, 80 );
  csput( 40, START_LINE + 2 * ( Boot->Sectors + 1 ) + 1,
	 "%c02Press %c0farrows%c02 to move "
	 "%c0fEnter%c02 to mark cluster as Bad "
	 "%c0fEsc%c02 to exit this mode" );
  do
  {
    i = sector / ( Boot->HeadCnt * Boot->Sectors );
    k = sector % ( Boot->HeadCnt * Boot->Sectors );
    j = k / Boot->Sectors;
    k %= Boot->Sectors;
    hilite( i + 1, START_LINE + j * ( Boot->Sectors + 1 ) + k, i + 1, revers );
    cput( 5, 1, START_LINE + 2 * Boot->Sectors + 1, '', 0x03, 80 );
    sprintf( line, " Absolute Sector %d Track %d Head %d Sector %d ",
	     sector, i, j, k + 1 );
    if ( sector >= ( Boot->ResSecs + Boot->FatSize * Boot->FatCnt +
		     ( Boot->RootSiz >> 4 ) ) )
    {
      cluster = 2 + ( sector - Boot->ResSecs - Boot->FatSize * Boot->FatCnt -
		      ( Boot->RootSiz >> 4 ) ) / Boot->ClustSiz;
      sprintf( line + strlen( line ), "Cluster %d ", cluster );
    }
    csput( 40, START_LINE + 2 * Boot->Sectors + 1, line );
    key = _bios_keybrd( _KEYBRD_READ );
    hilite( i + 1, START_LINE + j * ( Boot->Sectors + 1 ) + k, i + 1, revers );
    switch ( key )
    {
      case UP_ARROW:
	sector--;
	break;
      case DOWN_ARROW:
	sector++;
	break;
      case LEFT_ARROW:
	sector -= Boot->Sectors * Boot->HeadCnt;
	break;
      case RIGHT_ARROW:
	sector += Boot->Sectors * Boot->HeadCnt;
	break;
      case HOME_KEY:
	sector = 0;
	break;
      case END_KEY:
	sector = Boot->TotSecs - 1;
	break;
      case ENTER:
	break;
    }
    if ( sector < 0 )
      sector = 0;
    if ( sector >= Boot->TotSecs )
      sector = Boot->TotSecs - 1;
  } while ( key != ESC );
}

static void
split_sect( int *track, int *head, int *sector, int abs_sect )
{
  *sector = abs_sect;
  *track = *sector / ( Boot->HeadCnt * Boot->Sectors );
  *sector -= *track * Boot->HeadCnt * Boot->Sectors;
  *head = *sector / Boot->Sectors;
  *sector -= *head * Boot->Sectors;
}


static int
write_chunk( byte * buffer, int s_beg, int s_end )
{
  int		t[2], h[2], s[2];
  int		error;

  error = NONE;
  split_sect( &t[0], &h[0], &s[0], s_beg );
  split_sect( &t[1], &h[1], &s[1], s_end );
  if ( t[0] == t[1] && h[0] == h[1] )
  {
    if ( error = int13( WRITE, t[0], h[0], buffer, s[0] + 1,
			s[1] + 1 - s[0] ) )
      return error;
  }
  else
  {
    if ( error = int13( WRITE, t[0], h[0], buffer, s[0] + 1,
			Boot->Sectors - s[0] ) )
      return error;
    if ( error = int13( WRITE, t[1], h[1],
		      buffer + ( Boot->Sectors - s[0] ) * 0x200, 1, s[1] ) )
      return error;
  }
  return error;
}

int
new_boot_fat( byte * fat, int mode )
{
  int		i, error;
  int		sector, cluster, type;

  error = NONE;
  if ( error = int13( WRITE, 0, 0, ( byte far * ) boot_start, 1, 1 ) )
    return error;
  if ( !mode )
    memset( fat, 0, Boot->FatSize * 0x200 );
  *( byte * ) fat = Boot->Media;
  *( word * ) ( fat + 1 ) = 0xffff;
  for ( i = 0; i < Boot->TotSecs; i++ )
    if ( sect_status[i] != 0xffff )
      if ( ( sect_status[i] & ERROR ) == ERROR )
	if ( i >= ( Boot->ResSecs + Boot->FatSize * Boot->FatCnt +
		    ( Boot->RootSiz >> 4 ) ) )
	{
	  sector = i - Boot->ResSecs + Boot->FatSize * Boot->FatCnt +
		  ( Boot->RootSiz >> 4 );
	  cluster = 2 + sector / Boot->ClustSiz;
	  type = *( word * ) ( fat + cluster * 3 / 2 );
	  if ( cluster & 1 )
	  {
	    type &= 0x000f;
	    type |= 0xff70;
	  }
	  else
	  {
	    type &= 0xf000;
	    type |= 0x0ff7;
	  }
	  *( word * ) ( fat + cluster * 3 / 2 ) = type;
	}
  for ( i = 0; i < Boot->FatCnt; i++ )
    if ( error = write_chunk( fat, Boot->ResSecs + i * Boot->FatSize,
			   Boot->ResSecs + ( i + 1 ) * Boot->FatSize - 1 ) )
      return error;
  if ( !mode )
  {
    memset( trkbuf, 0, ( Boot->RootSiz >> 4 ) * 0x200 );
    if ( error = write_chunk( trkbuf, Boot->ResSecs +
			      Boot->FatCnt * Boot->FatSize,
			      Boot->ResSecs + Boot->FatCnt * Boot->FatSize +
			      ( Boot->RootSiz >> 4 ) - 1 ) )
      return error;
  }
  return error;
}

int
dos_recover(  )
{
  int		i, j, k, l[3], key;
  char	       *help_tbl[] = {
    "%c20B%c02 - Boot sector",
    "%c20F%c02 - First FAT",
    "%c20f%c02 - Second FAT",
    "%c20D%c02 - Root directory",
    "%c20d%c02 - other directory",
    "%c20r%c02 - reserved sectors",
    "%c20%c02 - used sector",
    "%c20.%c02 - free sector",
    "%c20X%c02 - Bad cluster"};

  if ( !map_active )
    swap_screen( 1 );
  map_active = 0;
  clear_screen( last_line + 1 );
  box( 10, 1, 70, 5, "ĳڿ", 0x03, ' ', 0x0a );
  csput( 40, 1, "%c0a Sectors legend: "
	 "(flashing means sector with an error) " );
  j = 11;
  for ( i = 0; i < sizeof( help_tbl ) / sizeof( char * ); i++ )
  {
    l[i % 3] = sput( j, 2 + ( i % 3 ), help_tbl[i] );
    if ( i % 3 == 3 - 1 )
    {
      for ( k = 1; k < 3; k++ )
	if ( l[k] > l[0] )
	  l[0] = l[k];
      j += 4 + l[0];
    }
  }
  show_map(  );
  k = last_line / 2;
  box( 10, k - 6, 70, k + 9, "ĳڿ", 0x03, ' ', 0x0a );

  csput( 40, k - 4, "%c2f Mark as Bad Cluster " );
  csput( 40, k - 2, "%c02If sector is part of a file then also move it" );
  csput( 40, k - 1, "to a sane place if extra space is available" );

  csput( 40, k + 1, "%c0aChose one of the following:" );

  sput( 13, k + 3, "%c0f1)%c02 all sectors with unrecoverable errors" );
  sput( 13, k + 4, "%c0f2)%c02 all sectors with recovered alignament and crc errors" );
  sput( 13, k + 5, "%c0f3)%c02 all sectors which had retries at read time" );
  sput( 13, k + 7, "%c0fM)%c02 manually mark as bad" );

  csput( 40, last_line - 2, "%c4f DOS recovery mode "
	 "not available yet in this version " );
  csput( 40, last_line - 1, "%c02 Press any key " );
  for ( ;; )
  {
    key = _bios_keybrd( _KEYBRD_READ );
    if ( key & 0xff )
      key &= 0xff;
    k = last_line / 2;
    box( 10, k - 6, 70, k + 9, "      ", 0x02, ' ', 0x02 );
    show_map(  );
    switch ( key )
    {
      case '1':
	break;
      case '2':
	break;
      case '3':
	break;
      case 'M':
      case 'm':
	position_sector(  );
	break;
      default:
	swap_screen( 0 );
	return 0;
    }
    cput( 5, 1, START_LINE + 2 * ( Boot->Sectors + 1 ) + 1, ' ', 0x02, 80 );
    csput( 40, START_LINE + 2 * ( Boot->Sectors + 1 ) + 1,
	   "%c02Choose: %c0f1%c02 errors "
	   "%c0f2%c02 recovered "
	   "%c0f3%c02 retried "
	   "%c0fM%c02 manually "
	   "%c0fEsc%c02 exit" );
  }
}
