#include <stdio.h>
#include <stdarg.h>
#include <conio.h>
#include <bios.h>
#include <string.h>
#include <malloc.h>

#include "revive.h"

static unsigned int
crc( char *buffer, word cmp_crc )
{
  int		i;
  word		crc;		/* Current CRC. */

  /*
   * This value coresponds to a starting value of -1 followed by 3 bytes of
   * 0xA1 and a 0xFB
   */
  crc = 0xE295;
  for ( i = 0; i < 0x200; i++ )
    crc = ( crc << 8 ) ^ crc_table[hib( crc ) ^ *( buffer + i )];
  return ( crc == ( cmp_crc << 8 | cmp_crc >> 8 ) );
}

static int
find_crc( char *buffer, word the_crc, int poz1, int poz2 )
{
  byte		orig1, orig2;
  byte	       *p1, *p2;
  int		found = 0;

  p1 = buffer + poz1;
  p2 = buffer + poz2;
  orig1 = *p1;
  orig2 = *p2;
  do
  {
    if ( crc( buffer, the_crc ) )
    {
      found = 1;
      break;
    }
    ( *p1 )++;
    if ( *p1 == orig1 )
      if ( poz1 == poz2 )
	break;
      else
	( *p2 )++;
  } while ( !( *p1 == orig1 && *p2 == orig2 ) );
  return found;
}

static int
try_restore( int track, int head )
{
  int		c_sect;
  int		s_crc[3][32];
  int		mem_err[7];
  byte		value[7];
  int		retry, read_error;
  int		i, j, k;
  int		count, first_byte, last_byte;
  int		return_val;
  word		some_crc, xcrc;

  for ( i = 0; i < Boot->Sectors; i++ )
  {
    retry = 0;
    do
    {
      while ( ( read_error = int13( READ, track, head,
				    trkbuf + i * 0x200, i + 1, 1 ) ) == 6 )
	;			/* ignore "disk changed" error */
      retry++;
    } while ( read_error && retry < 3 );
    if ( read_error )
    {
      SECT_STATUS( track, head, i ) = READ | ERROR;
      verbose( track, head, i + 1, 'E', "read" );
    }
    else
    {
      SECT_STATUS( track, head, i ) = READ | RECOVER | RETRY;
      /* verbose( track, head, i + 1, 'R', "retry" ); */
    }
  }
  if ( intlv_one )
  {
    for ( j = 0; j < 3; j++ )
    {
      for ( i = 0; i < Boot->Sectors; i++ )
	s_crc[j][i] = 0;
      if ( int13( READ_TRACK, track, head, trkbuf + ( j + 1 ) * 0x4000, 0, 0 ) )
	return 0;
    }
    c_sect = 0;
    while ( 1 )
    {
      if ( ( ERROR & SECT_STATUS( track, head, c_sect ) ) == ERROR )
      {
	for ( j = 0; j < 3; j++ )
	  /* are these sectors aligned to mark ? */
	  if ( c_sect == 0 || s_crc[j][c_sect] )
	    if ( crc( trkbuf + ( j + 1 ) * 0x4000 + c_sect * 0x200,
		      *( word * ) ( trkbuf + ( j + 1 ) * 0x4000 + ( c_sect + 1 ) * 0x200 ) ) )
	    {
	      memcpy( trkbuf + c_sect * 0x200,
		      trkbuf + ( j + 1 ) * 0x4000 + c_sect * 0x200, 0x200 );
	      SECT_STATUS( track, head, c_sect ) = READ | RECOVER | ALIGN;
	      verbose( track, head, c_sect + 1, 'R', "re-aligned" );
	      break;
	    }
	if ( j != 3 )
	  continue;
	count = 0;
	/* read (with error) 4 times the sector */
	for ( i = 0; i < 4; i++ )
	{
	  mem_err[i + 3] = int13( READ, track, head,
			  trkbuf + ( i * 0x4000 ) + 0x3e00, c_sect + 1, 1 );
	  /* only CRC error (or no errror) reads can be useful */
	  if ( mem_err[i + 3] == 0x10 || mem_err[i + 3] == 0 )
	  {
	    mem_err[i + 3] = 1;
	    count++;
	  }
	  else
	    mem_err[i + 3] = 0;
	}
	for ( j = 0; j < 3; j++ )
	{
	  mem_err[j] = 0;
	  if ( c_sect == 0 || s_crc[j][c_sect] )
	  {
	    if ( *( trkbuf + ( j + 1 ) * 0x4000 + ( c_sect + 1 ) *
		    0x200 + 2 ) == 0x4e )
	    {
	      mem_err[j] = 1;
	      count++;
	    }
	  }
	}
	switch ( count )
	{
	  case 0:
	    break;
	  case 1:
	    j = 0;
	    while ( !mem_err[j] )
	      j++;
	    if ( j < 3 )
	      memcpy( trkbuf + c_sect * 0x200,
		      trkbuf + ( j + 1 ) * 0x4000 + c_sect * 0x200, 0x200 );
	    else
	      memcpy( trkbuf + c_sect * 0x200,
		      trkbuf + ( j - 3 ) * 0x4000 + 0x3e00, 0x200 );
	    break;
	  default:
	    first_byte = 0;
	    last_byte = 0;
	    for ( i = 0; i < 0x200; i++ )
	    {
	      k = 0;
	      for ( j = 0; j < 7; j++ )
		if ( mem_err[j] )
		  value[k++] = ( j < 3 ?
		     *( trkbuf + ( j + 1 ) * 0x4000 + c_sect * 0x200 + i ) :
			    *( trkbuf + ( j - 3 ) * 0x4000 + 0x3e00 + i ) );
	      xcrc = med_byte( &value, count );
	      if ( ( xcrc & 0xff00 ) != 0 )
		if ( !first_byte )
		  first_byte = i;
		else
		  last_byte = i;
	      *( trkbuf + c_sect * 0x200 + i ) = ( byte ) xcrc;
	    }
	    for ( j = 0; j < 3; j++ )
	      if ( mem_err[j] )
	      {
		some_crc = *( word * ) ( trkbuf + ( j + 1 ) * 0x4000 +
					 ( c_sect + 1 ) * 0x200 );
		break;
	      }
	    if ( j != 3 )
	      if ( find_crc( trkbuf + c_sect * 0x200,
			     some_crc, first_byte, last_byte ) )
	      {
		SECT_STATUS( track, head, c_sect ) = READ | RECOVER | CRC;
		verbose( track, head, c_sect + 1, 'R', "crc" );
	      }
	    break;
	}
      }
      if ( c_sect == Boot->Sectors - 1 )
	break;
      for ( j = 0; j < 3; j++ )
	s_crc[j][c_sect + 1] =
		re_align( trkbuf + ( j + 1 ) * 0x4000 + ( c_sect + 1 ) * 0x200 );
      c_sect++;
    }
  }
  return_val = 1;
  for ( i = 0; i < Boot->Sectors; i++ )
  {
    switch ( SECT_STATUS( track, head, i ) & 0xff00 )
    {
      case ERROR:
      case RECOVER | ALIGN:
      case RECOVER | CRC:
	SD[file_secs] = malloc( sizeof( sector_s ) );
	SD[file_secs]->track = track;
	SD[file_secs]->head = head;
	SD[file_secs]->sect = i;
	SD[file_secs]->abs_sect = ( track * Boot->HeadCnt + head ) * Boot->Sectors + i;
	SD[file_secs]->error_type = SECT_STATUS( track, head, i );
	fseek( sb, ( unsigned long ) sizeof( int ) + file_secs * 0x200,
	       SEEK_SET );
	fwrite( trkbuf + i * 0x200, 0x200, 1, sb );
	file_secs++;
	break;
      default:
	break;
    }
    if ( ( ERROR & SECT_STATUS( track, head, i ) ) == ERROR )
      return_val = 0;
  }
  return return_val;
}

static void
verbose_mode( int mode )
{
  switch ( mode & 0xff )
  {
    case NONE:
      cput( 6, 76, END_LINE - 2, '', 0x04, 13, 2 );
      break;
    case READ:
      esput( 76, END_LINE - 2, "%c0e Quick mode " );
      break;
    case FORMAT:
      esput( 76, END_LINE - 2, "%c0e Format mode " );
      break;
  }
}

static void
help_line( int str_nr )
{
  cput( 5, 1, last_line, ' ', 2, 80 - strlen( "Esc - EXIT" ) );
  switch ( str_nr )
  {
    case 0:
      sprintf( line, "%%c0fPress %%c0e'F'%%c02 - to toggle format mode, "
	       "%%c0e'Q'%%c02 - to toggle quick mode" );
      break;
    case 1:
      sprintf( line, "%%c8cError found%%c02 press "
	       "%%c0e'S'%%c02 - to Skip, "
	       "%%c0e'R'%%c02 - to Recover, "
	       "%%c0e'A'%%c02 - for Auto" );
      break;
  }
  sput( 1, last_line, line );
}

int
revive( void )
{
  int		i, t, h, s;
  int		second_pass, retry, key;
  int		read_error, verify_error;
  int		format_error, write_error;
  int		temp_fresh_boot;
  unsigned long id;
  unsigned int	status, mode;
  char	       *p;

  mode = NONE;
  temp_fresh_boot = 0;
  cput( 5, 1, last_line, ' ', 2, 80 - strlen( "Esc - EXIT" ) );
  sprintf( line, "%%c0eAny key %%c02 - to proceed, "
	   "%%c0e'F'%%c02 - format, %%c0e'Q'%%c02 - quick" );
  sput( 1, last_line, line );
  key = _bios_keybrd( _KEYBRD_READ );
  if ( ( key & 0xdf ) == 'F' )
    temp_fresh_boot = 1;
  cput( 5, 2, END_LINE, '', 0x04, 78 );
  help_line( 0 );
  if ( key == ESC )
    return NONE;
  for ( t = 0; t < Tracks; t++ )
    for ( h = 0; h < Boot->HeadCnt; h++ )
    {
      second_pass = 0;
      do
      {
	status = NONE;
	if ( key || _bios_keybrd( _KEYBRD_READY ) )
	{
	  if ( !key )
	    key = _bios_keybrd( _KEYBRD_READ );
	  if ( key == ESC )
	  {
	    if ( fresh_boot || temp_fresh_boot )
	      new_boot_fat( trkbuf, 0 );
	    return NONE;
	  }
	  key &= 0xdf;
	  if ( key == 'F' )
	  {
	    if ( ( mode & 0xff ) == FORMAT )
	    {
	      mode = NONE;
	    }
	    else
	    {
	      mode = FORMAT;
	      _fmemset( trkbuf, 0xf6, Boot->Sectors * Boot->SectSiz );
	    }
	    verbose_mode( mode );
	  }
	  if ( key == 'Q' )
	  {
	    if ( ( mode & 0xff ) == READ )
	      mode = NONE;
	    else
	      mode = READ | RETRY;
	    verbose_mode( mode );
	  }
	  key = 0;
	}
	if ( mode != FORMAT )
	{
	  retry = 0;
	  do
	  {
	    while ( ( read_error = int13( READ, t, h, trkbuf, 1, Boot->Sectors ) ) == 6 )
	      ;			/* ignore "disk changed" error */
	    retry++;
	  } while ( read_error && retry < 3 );
	  if ( read_error )
	  {
	    cput( 5, 2, BAR_LINE + 2, '', 0x03, Boot->Sectors + 13 );
	    sput( Boot->Sectors + 3, BAR_LINE + 2, "%c03sector ID's " );
	    for ( i = 0; i < Boot->Sectors; i++ )
	    {
	      int13( READ_ID, &id );
	      cput( 4, 1 + ( byte ) ( id >> 16 ), BAR_LINE + 2, '' - h,
		    0x03 );
	    }
	    status = ERROR;
	    mark_active( t, h, status );
	    if ( try_restore( t, h ) )
	    {
	      status = RECOVER;
	      mark_active( t, h, status );
	    }
	    cput( 5, 2, BAR_LINE + 2, '', 0x03, Boot->Sectors + 13 );
	  }
	  else
	  {
	    if ( !second_pass )
	    {
	      for ( i = 0; i < Boot->Sectors; i++ )
		SECT_STATUS( t, h, i ) = READ;
	      status = READ;
	      mark_active( t, h, status );
	    }
	  }
	}
	if ( read_error )
	{
	  if ( mode == READ )
	    continue;
	  if ( mode == ( READ | RETRY ) )
	  {
	    help_line( 1 );
	    do
	    {
	      key = _bios_keybrd( _KEYBRD_READ );
	      if ( key == ESC )
		break;
	      switch ( key & 0xdf )
	      {
		case 'S':
		case 'R':
		  break;
		case 'Q':
		  mode = NONE;
		  verbose_mode(mode);
		  break;
		case 'F':
		  mode = FORMAT;
		  verbose_mode(mode);
		  break;
		case 'A':
		  mode = READ | ERROR;
		  break;
		default:
		  key = 0;
	      }
	    } while ( key == 0 );
	  }
	  help_line( 0 );
	  if ( key == ESC )
	    continue;
	  key &= 0xdf;
	  if ( key == 'S' )
	  {
	    key = 0;
	    continue;
	  }
	  key = 0;
	}
	if ( ( mode & 0xff ) != READ || read_error )
	{
	  p = fmtbuf;
	  for ( s = 1; s <= Boot->Sectors; s++ )
	  {
	    *p++ = t;
	    *p++ = h;
	    *p++ = s;
	    *p++ = Boot->SectSiz >> 8;
	  }
	  retry = 0;
	  do
	  {
	    format_error = int13( FORMAT, t, h, ( char * ) fmtbuf,
				  1, Boot->Sectors );
	    if ( format_error == 3 )	/* write protected */
	      return FORMAT;
	    if ( mode == FORMAT )
	    {
	      for ( i = 0; i < Boot->Sectors; i++ )
		SECT_STATUS( t, h, i ) = FORMAT << 8;
	    }
	    else
	    {
	      write_error = int13( WRITE, t, h, trkbuf, 1, Boot->Sectors );
	      if ( format_error || write_error )
	      {
		/* write as many sectors as possible in case of error */
		for ( i = 0; i < Boot->Sectors; i++ )
		  SECT_STATUS( t, h, i ) = WRITE |
			  ( int13( WRITE, t, h, trkbuf + i * 0x200,
				   i + 1, 1 ) ? ERROR : NONE );
		status = ERROR;
		mark_active( t, h, status );
		break;
	      }
	    }
	    verify_error = int13( VERIFY, t, h, trkbuf, 1, Boot->Sectors );
	    retry++;
	  } while ( verify_error && retry < 5 );
	  if ( retry == 5 )
	  {
	    status = ERROR;
	    mark_active( t, h, status );
	  }
	  second_pass = 1;
	}
      } while ( status == RECOVER );
    }
  if ( fresh_boot || temp_fresh_boot )
    new_boot_fat( trkbuf, 0 );
  return NONE;
}
