/* revive (put new format info) a floppy disk */

#include <stdlib.h>
#include <stdio.h>
#include <conio.h>
#include <io.h>
#include <dos.h>
#include <bios.h>
#include <string.h>
#include <malloc.h>

#include "revive.h"

struct Boot_t *Boot;
int	      Tracks;
byte	      Drive;
word	      crc_table[256];	/* CRC table. */
char	     *line, *fmtbuf;
byte	     *buf128, *trkbuf;
byte	     *fat;
int	      fresh_boot;
unsigned int *sect_status;
word	     *scr_zone;
int	      current_ch;
int	      otrk;
int	      screen_poz;
int	      last_line;
int	      map_active;
int	      intlv_one;
sector_s     *SD[MAX_SECTORS];
int	      file_secs;
FILE	     *sb;

static void
init_crc(  )
{
  int		i;
  word	       *halfi;		/* Pointer to CRC of i / 2. */
  const		POLYNOM = 0x1021;

  /* Generate a CRC lookup table for faster calculation. */
  for ( i = 0, halfi = crc_table, crc_table[0] = 0;
	i < 256; i += 2, halfi++ )
    if ( hib( *halfi ) & 0x80 )
      crc_table[i] = ( crc_table[i + 1] = *halfi << 1 ) ^ POLYNOM;
    else
      crc_table[i + 1] = ( crc_table[i] = *halfi << 1 ) ^ POLYNOM;
}

static void
usage( char *progname )
{
  fprintf( stderr, "\n\nUsage:\t%s [-tsb] drive\n", progname );
  fprintf( stderr, "\t-t:nn\tnumber of tracks\n" );
  fprintf( stderr, "\t-s:nn\tnumber of sectors\n" );
  fprintf( stderr, "\t-b\tput fresh boot/fat info\n" );
  exit( 0 );
}

/* main program */
/* revive a floppy disk */
int
main( int argc, char *argv[] )
{
  int		i, j, k, key;
  int		error_code, valid_boot;
  char	       *p;
  int		supp_trk[] = {40, 80};
  int		supp_sect[] = {8, 9, 15, 18};
  char	       *revive_name = "revive.dat";
  char	       *progname;

  progname = malloc( 80 * sizeof( char ) );
  _splitpath( argv[0], NULL, NULL, progname, NULL );
  Boot = ( struct Boot_t * ) ( ( byte far * ) boot_start + 11 );
  fresh_boot = 0;
  if ( argc > 1 )
  {
    while ( --argc )
    {
      p = argv[argc];
      if ( *p == '/' || *p == '-' )
      {
	*++p |= 0x20;
	switch ( *p++ )
	{
	  case 't':
	    if ( *p++ != ':' )
	      usage( progname );
	    if ( sscanf( p, "%d", &Tracks ) != 1 )
	      usage( progname );
	    break;
	  case 's':
	    if ( *p++ != ':' )
	      usage( progname );
	    if ( sscanf( p, "%d", &Boot->Sectors ) != 1 )
	      usage( progname );
	    break;
	  case 'b':
	    fresh_boot = 1;
	    break;
	  case '?':
	  case 'h':
	  default:
	    usage( progname );
	}
      }
      else
      {
	*p |= 0x20;
	if ( ( Drive = *p - 'a' ) > 1 )
	  usage( progname );
      }
    }
  }
  else
    usage( progname );

  line = malloc( 160 );
  fmtbuf = malloc( 4 * 32 );
  buf128 = halloc( 0x20000L, sizeof( byte ) );	/* allocate 2 * 64K segments */
  if ( buf128 == NULL )
  {
    fprintf( stderr, "Cannot allocate 128k buffer\n" );
    exit( -1 );
  }
  /* align to 64k boundary */
  trkbuf = MK_FP( ( ( FP_SEG( buf128 ) + 0xfff ) & 0xf000 ), 0 );
  save_video_mode(  );
  hook_int_0E( );
  do
  {
    intlv_one = 1;
    Tracks = 0;
    Boot->Sectors = 0;
    last_line = set_hi_text_mode( 0 );
    hide_cursor(  );
    box( 1, BAR_LINE - 1, 80, BAR_LINE + 2, "ĳڿ", 0x03, ' ', 0x0a );
    box( 1, START_LINE - 1, 80, END_LINE, "ͺɻȼ", 0x4, ' ', 0x0a );
    cput( 5, 2, START_LINE, '', 4, 78 );
    cput( 4, 1, START_LINE, '', 4 );
    cput( 4, 80, START_LINE, '', 4 );
    esput( 80, last_line, "%c0eEsc%c02 - EXIT" );
    csput( 40, 1, "%c1f Revive disk v1.0 (C) EmilSoft 1998 " );
    current_ch = '' + 0x200;
    otrk = 0;
    show_tracks(  );
    esput( 75, START_LINE, " %c0fR%c02 recovered %c04 %c0cE%c02 error " );
    cput( 5, 2, END_LINE - 2, '', 4, 78 );
    cput( 4, 1, END_LINE - 2, '', 4 );
    cput( 4, 80, END_LINE - 2, '', 4 );
    for ( i = 0; i < S_COLS; i++ )
      sput( 2 + ( 76 / S_COLS ) * i, END_LINE - 2, "%c04 t/h/s " );
    csput( 40, END_LINE - 2, "%c0F Status " );
    sprintf( line, "%%c04 Interrupt 13 found at address %FP ", true13(  ) );
    sput( 2, END_LINE, line );

    do
    {
      error_code = 1;
      for ( i = sizeof( supp_trk ) / sizeof( int ); i > 0; i-- )
	for ( j = sizeof( supp_sect ) / sizeof( int ); j > 0; j-- )
	  if ( error_code )
	  {
	    int13( SET_MEDIA, supp_trk[i - 1], supp_sect[j - 1] );
	    int13( RESET );
	    error_code = int13( READ_TRACK, 0, 0, trkbuf, 0, 0 );
	  }
	  else
	    break;
      intlv_one = interleave_one( trkbuf );
      error_code = int13( READ, 0, 0, trkbuf, 1, 1 );
      if ( error_code != 0 && error_code != 6 )
      {
	cput( 5, 1, last_line, ' ', 2, 80 );
	sput( 1, last_line, "%c0eAny key%c02 - to try again or "
	      "%c0eESC%c02 - to skip" );
	key = _bios_keybrd( _KEYBRD_READ );
	cput( 5, 1, last_line, ' ', 2, 80 );
	esput( 80, last_line, "%c0eEsc%c02 - EXIT" );
      }
    }
    while ( error_code != 0 && key != ESC );

    sprintf( line, "%%c04 Interleave is%s 1 ", intlv_one ? "" : " not" );
    esput( 79, END_LINE, line );
    scr_zone = NULL;
    sect_status = NULL;
    map_active = 0;
    if ( key != ESC && ( *trkbuf == 0xeb || *trkbuf == 0xe9 ) )
    {
      Boot->SectSiz = *( int * ) ( trkbuf + 0x0b );	/* bytes per sector */
      Boot->ClustSiz = *( trkbuf + 0x0d );	/* sectors per cluster */
      Boot->ResSecs = *( int * ) ( trkbuf + 0x0e );	/* reserved sectors
							 * (before FAT) */
      Boot->FatCnt = *( trkbuf + 0x10 );	/* number of FATs */
      Boot->RootSiz = *( int * ) ( trkbuf + 0x11 );	/* maximum root
							 * directory entries */
      Boot->TotSecs = *( int * ) ( trkbuf + 0x13 );	/* total number of
							 * sectors in Media */
      Boot->Media = *( trkbuf + 0x15 ); /* Media descriptor */
      Boot->FatSize = *( int * ) ( trkbuf + 0x16 );	/* number of sectors in
							 * one FAT */
      Boot->Sectors = *( int * ) ( trkbuf + 0x18 );	/* sectors per track */
      Boot->HeadCnt = *( int * ) ( trkbuf + 0x1a );	/* number of read/write
							 * heads */
      Boot->HidSectL = *( int * ) ( trkbuf + 0x1b );	/* hidden sectors */
      Tracks = Boot->TotSecs / Boot->Sectors / Boot->HeadCnt;
      valid_boot = 1;
    }
    if ( Boot->SectSiz != 0x200 )
    {
      Boot->SectSiz = 0x200;
      valid_boot = 0;
    }
    if ( Boot->HeadCnt < 1 || Boot->HeadCnt > 2 )
    {
      Boot->HeadCnt = 2;
      valid_boot = 0;
    }
    if ( Boot->ClustSiz == 0 )
    {
      Boot->ClustSiz = 1;
      valid_boot = 0;
    }
    if ( Boot->Media == 0 )
    {
      Boot->Media = 0xf0;
      valid_boot = 0;
    }
    if ( Tracks <= 0 || Tracks > 80 )
    {
      Tracks = 80;
      valid_boot = 0;
    }
    if ( Boot->Sectors <= 0 || Boot->Sectors > 18 )
    {
      Boot->Sectors = 18;
      valid_boot = 0;
    }
    sprintf( line, "%%c04 Drive %%c0c%c: %%c04type:%%c0c%02X "
	     "%%c04tracks:%%c0c%d %%c04sectors:%%c0c%d "
	     "%%c04SectSize:%%c0c%d %%c04clustsize:%%c0c%d "
	     "%%c04heads:%%c0c%d ",
	     'A' + Drive, Boot->Media, Tracks, Boot->Sectors,
	     Boot->SectSiz, Boot->ClustSiz, Boot->HeadCnt );
    sput( 4, START_LINE - 1, line );
    int13( SET_MEDIA, Tracks, Boot->Sectors );
    int13( RESET );
    Boot->TotSecs = Tracks * Boot->HeadCnt * Boot->Sectors;
    if ( ( sect_status = ( unsigned int * ) malloc( Boot->TotSecs *
					sizeof( unsigned int ) ) ) == NULL )
    {
      restore_video_mode(  );
      fprintf( stderr, "Allocation error\n" );
      return -1;
    }
    memset( sect_status, 0xff, Boot->TotSecs * sizeof( unsigned int ) );
    init_crc(  );
    screen_poz = 0;
    sb = fopen( revive_name, "wb" );
    if ( sb != NULL )
    {
      file_secs = 0;
      memset( &SD, 0, MAX_SECTORS * sizeof( sector_s * ) );
    }
    /* main revive function */
    error_code = revive(  );
    if ( sb != NULL )
    {
      fseek( sb, 0L, SEEK_SET );
      fwrite( &file_secs, sizeof( int ), 1, sb );
      fseek( sb, 0L, SEEK_END );
      for ( i = 0; i < file_secs; i++ )
      {
	fwrite( SD[i], sizeof( sector_s ), 1, sb );
	free( SD[i] );
      }
      fclose( sb );
      sb = NULL;
    }
    cput( 5, 2, END_LINE - 1, ' ', 0, 78 );
    switch ( error_code )
    {
      case NONE:
	csput( 40, END_LINE - 1, "%c0fRevive finished" );
	break;
      case FORMAT:
	csput( 40, END_LINE - 1, "%c4F write-protected disk ... "
	       "unprotect then try again " );
	break;
    }
    do
    {
      sput( 1, last_line, "%c02Press "
	    "%c0f'N'%c02 for next disk %c06: "
	    "%c0f'M'%c02 to toggle map %c06: "
	    "%c01'D'%c02 for dos recover" );
      key = _bios_keybrd( _KEYBRD_READ );
      key &= 0xdf;
      if ( key == 'M' )
	toggle_map(  );
      if ( key == 'D' )
      {
	if ( valid_boot )
	{
	  fat = malloc( Boot->FatSize * 0x200 );
	  for ( i = 0; i < Boot->FatSize; i++ )
	  {
	    k = Boot->ResSecs + i;
	    j = k / Boot->Sectors;
	    k %= Boot->Sectors;
	    k = int13( READ, 0, j, fat + i * 0x200, k + 1, 1 );
	    if ( k )		/* error try second fat */
	    {
	      k = Boot->ResSecs + Boot->FatSize + i;
	      j = k / Boot->Sectors;
	      k %= Boot->Sectors;
	      k = int13( READ, 0, j, fat + i * 0x200, k + 1, 1 );
	    }
	    if ( k )
	      break;
	  }
	  if ( !k )
	    dos_recover(  );
	  free( fat );
	}
      }
    } while ( key == 'M' || key == 'D' );
    if ( scr_zone != NULL )
      free( scr_zone );
    if ( sect_status != NULL )
      free( sect_status );
  } while ( key == 'N' );
  unlink( revive_name );
  restore_int_0E( );
  restore_video_mode(  );
  hfree( buf128 );
  free( fmtbuf );
  free( line );
  return 0;
}
